
(** testing infrastructure.  See .mli for description *)

module type Examples = sig
  type t
  val examples : t list
end


let uncurry_1 f a       = f a
let uncurry_2 f (a,b)   = f a b
let uncurry_3 f (a,b,c) = f a b c

(** returns a list of all (name, value) pairs, where name is something like "input 1" *)
let named_singles lst =
  List.mapi (fun ia a -> "input "^string_of_int ia, a) lst

(** returns a list of all (name, (v1,v2)) pairs, where name is something like "inputs 1 and 2" *)
let named_pairs lst1 lst2 =
  let name ia ib = "inputs "^string_of_int ia^" and "^string_of_int ib in
  List.concat (List.mapi (fun ia a -> List.mapi (fun ib b -> name ia ib, (a,b)) lst2) lst1)

(** returns a list of all (name, (v1,v2,v3)) pairs, where name is something like "inputs 1 2 and 3" *)
let named_triples lst1 lst2 lst3 =
  let name ia ib ic = "inputs "^string_of_int ia^" "^string_of_int ib^" and "^string_of_int ic in
  List.concat (List.mapi (fun ia a ->
    List.concat (List.mapi (fun ib b ->
                   List.mapi (fun ic c ->
                     name ia ib ic, (a,b,c)
                   ) lst3
                 ) lst2
               )) lst3)

exception SpecFailure of string

(** Raises a SpecFailure exception describing the failed inputs for the given
    test if the test does in fact fail; returns unit if the test does not fail. *)
let check named_inputs test =
  let failures  = List.filter (fun (name,input) -> try not (test input) with _ -> true) named_inputs in
  let strings   = List.map fst failures in
  let shortened = if List.length strings <= 3 then strings else
  		[List.nth strings 0; List.nth strings 1; List.nth strings 2;
  		"... and "^(string_of_int (List.length strings - 3))^" others"] in
  let msg       = "test failed on "^(String.concat ", " shortened) in
  if failures <> [] then raise (SpecFailure msg)

(* raises a SpecFailure if test fails on any of the inputs *)
let check1 inputs test =
  check (named_singles inputs) test

(* raises a SpecFailure if test fails on any pair of inputs *)
let check2 inputs test =
  check (named_pairs inputs inputs) (uncurry_2 test)

(* raises a SpecFailure if test fails on any triple of inputs *)
let check3 inputs test =
  check (named_triples inputs inputs inputs) (uncurry_3 test)

